/*
 * dbook.c --
 *
 *	Definition for packer wrappers to the pilot addressbook.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: abook.c,v 1.1 1999/03/27 13:26:31 aku Exp $
 */

#include "wrap.h"
#include "dates.h"
#include "category.h"
#include "internal.h"

#include "pi-address.h"

typedef struct Address         Address;
typedef struct AddressAppInfo  AddressAppInfo;


/*
 * Declarations of internal custom converter procedures.
 */

static Tcl_Obj*
LToObj _ANSI_ARGS_ ((Tcl_Interp* interp,
		     void*       itemPtr,
		     ClientData  clientData));

static int
LFromObj _ANSI_ARGS_ ((Tcl_Interp* interp,
		       Tcl_Obj*    o,
		       void*       itemPtr,
		       ClientData  clientData));

static Tcl_Obj*
PToObj _ANSI_ARGS_ ((Tcl_Interp* interp,
		     void*       itemPtr,
		     ClientData  clientData));

static int
PFromObj _ANSI_ARGS_ ((Tcl_Interp* interp,
		       Tcl_Obj*    o,
		       void*       itemPtr,
		       ClientData  clientData));

/*
 * Definitions for conversion of todo records.
 */

static TclUtil_EnumItem labelValues [] = {
  {"lastName",  0},  {"firstName", 1},  {"company",   2},  {"phone1",    3},
  {"phone2",    4},  {"phone3",    5},  {"phone4",    6},  {"phone5",    7},
  {"address",   8},  {"city",      9},  {"state",    10},  {"zipcode",  11},
  {"country",  12},  {"title",    13},  {"custom1",  14},  {"custom2",  15},
  {"custom3",  16},  {"custom4",  17},  {"note",     18},  {"phone6",   19},
  {"phone7",   20},  {"phone8",   21},
  {NULL,      -1},
};

static TclUtil_Enum labelEnum = {
  "label keys", NULL, labelValues
};


static TclUtil_StructItem phoneItems [] = {
  ITEM_CUSTOM ("A", phoneLabel [0], Address, PToObj, PFromObj, NULL),
  ITEM_CUSTOM ("B", phoneLabel [1], Address, PToObj, PFromObj, NULL),
  ITEM_CUSTOM ("C", phoneLabel [2], Address, PToObj, PFromObj, NULL),
  ITEM_CUSTOM ("D", phoneLabel [3], Address, PToObj, PFromObj, NULL),
  ITEM_CUSTOM ("E", phoneLabel [4], Address, PToObj, PFromObj, NULL),
  ITEM_END
};

static TclUtil_Struct phoneStruct = {
  "phoneLabels", 0, NULL, phoneItems
};


static TclUtil_StructItem recordItems [] = {
  ITEM        ("showPhone",   showPhone,  Address, TUSI_Boolean),
  ITEM_STRUCT ("phoneLabels", phoneLabel, Address, &phoneStruct),
  ITEM        ("lastName",    entry [ 0], Address, TUSI_String),
  ITEM        ("firstName",   entry [ 1], Address, TUSI_String),
  ITEM        ("company",     entry [ 2], Address, TUSI_String),
  ITEM        ("phoneA",      entry [ 3], Address, TUSI_String),
  ITEM        ("phoneB",      entry [ 4], Address, TUSI_String),
  ITEM        ("phoneC",      entry [ 5], Address, TUSI_String),
  ITEM        ("phoneD",      entry [ 6], Address, TUSI_String),
  ITEM        ("phoneE",      entry [ 7], Address, TUSI_String),
  ITEM        ("address",     entry [ 8], Address, TUSI_String),
  ITEM        ("city",        entry [ 9], Address, TUSI_String),
  ITEM        ("state",       entry [10], Address, TUSI_String),
  ITEM        ("zipcode",     entry [11], Address, TUSI_String),
  ITEM        ("country",     entry [12], Address, TUSI_String),
  ITEM        ("title",       entry [13], Address, TUSI_String),
  ITEM        ("custom1",     entry [14], Address, TUSI_String),
  ITEM        ("custom2",     entry [15], Address, TUSI_String),
  ITEM        ("custom3",     entry [16], Address, TUSI_String),
  ITEM        ("custom4",     entry [17], Address, TUSI_String),
  ITEM        ("note",        entry [18], Address, TUSI_String),
  ITEM_END
};

static TclUtil_Struct recordStruct = {
  "Address", 0, NULL, recordItems
};

static PiTcl_CvtData recordConverter = {
  &recordStruct,                     /* Definition of Address structure, for
				      * conversion to tcl list */
  sizeof (Address),                  /* Size of the Address structure */
  (PiTcl_FreeStruct*) free_Address,  /* Proc to free additionally allocated
				      * space */
  (PiTcl_CvtStruct*) pack_Address,   /* Serialize Address into bytestream */
  (PiTcl_CvtStruct*) unpack_Address, /* Unserialize bytestream into an Address
				      */
  4096                               /* Initial size of buffer to use for
				      * serialization */
};


/*
 * Definitions for conversion of Address application information.
 */

static TclUtil_StructItem appinfoItems [] = {
  ITEM_STRUCT ("category",      category,      AddressAppInfo, &catStruct),
  ITEM        ("country",       country,       AddressAppInfo, TUSI_Int),
  ITEM        ("sortByCompany", sortByCompany, AddressAppInfo, TUSI_Boolean),
  ITEM_CUSTOM ("labels",        category,      AddressAppInfo, LToObj, LFromObj, NULL),
  ITEM_END
};

static TclUtil_Struct appinfoStruct = {
  "Address application information", 0, NULL, appinfoItems
};


static PiTcl_CvtData appinfoConverter = {
  &appinfoStruct,                           /* Definition of Address appinfo
					     * structure, for conversion to
					     * and from a tcl list */
  sizeof (AddressAppInfo),                  /* Size of the Address appinfo
					     * structure */
  NULL,                                     /* *No* additionally allocated
					     * space */
  (PiTcl_CvtStruct*) pack_AddressAppInfo,   /* Serialize Address appinfo into
					     * a bytestream */
  (PiTcl_CvtStruct*) unpack_AddressAppInfo, /* Unserialize bytestream into an
					     * Address appinfo block */
  sizeof (AddressAppInfo)+1                 /* Initial size of buffer to use
					     * for serialization */
};

/*
 * Address databases do not have a sort information block.
 */

/*
 * Definition of the whole Address converter.
 */

static PiTcl_PilotCvt converter = {
  &recordConverter,
  NULL,
  &appinfoConverter
};

/*
 *------------------------------------------------------*
 *
 *	PiTclInitAddress --
 *
 *	------------------------------------------------*
 *	This procedure registers the commands wrapped
 *	around the (un)pack functionality for Addresss.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclInitAddress (interp)
Tcl_Interp* interp;
{
  Tcl_CreateObjCommand (interp, "piAddress",
			&PiTclPilotCvtCmd,
			(ClientData) &converter,
			NULL);

  return TCL_OK;
}



static Tcl_Obj*
LToObj (interp, itemPtr, clientData)
Tcl_Interp* interp;
void*       itemPtr;
ClientData  clientData;
{
  AddressAppInfo* a    = (AddressAppInfo*) itemPtr;
  Tcl_Obj*        list = Tcl_NewListObj (0, NULL);
  Tcl_Obj*        o;
  Tcl_Obj*        o2;
  int             i, res;

  if (list == (Tcl_Obj*) NULL) {
    return (Tcl_Obj*) NULL;
  }

  for (i=0; i < 22; i++) {

    o = TclUtil_EnumDecode (interp, &labelEnum, i);

    if (o == (Tcl_Obj*) NULL) {
      Tcl_DecrRefCount (list);
      return (Tcl_Obj*) NULL;
    }

    res = Tcl_ListObjAppendElement (interp, list, o);

    if (res != TCL_OK) {
      Tcl_DecrRefCount (list);
      return (Tcl_Obj*) NULL;
    }

    o = Tcl_NewListObj (0, NULL);

    if (o == (Tcl_Obj*) NULL) {
      Tcl_DecrRefCount (list);
      return (Tcl_Obj*) NULL;
    }

    
    o2 = Tcl_NewIntObj (a->labelRenamed [i]);

    if (o2 == (Tcl_Obj*) NULL) {
      Tcl_DecrRefCount (list);
      Tcl_DecrRefCount (o);
      return (Tcl_Obj*) NULL;
    }

    res = Tcl_ListObjAppendElement (interp, o, o2);

    if (res != TCL_OK) {
      Tcl_DecrRefCount (list);
      Tcl_DecrRefCount (o);
      return (Tcl_Obj*) NULL;
    }

    o2 = Tcl_NewStringObj (a->labels [i], -1);

    if (o2 == (Tcl_Obj*) NULL) {
      Tcl_DecrRefCount (list);
      Tcl_DecrRefCount (o);
      return (Tcl_Obj*) NULL;
    }

    res = Tcl_ListObjAppendElement (interp, o, o2);

    if (res != TCL_OK) {
      Tcl_DecrRefCount (list);
      Tcl_DecrRefCount (o);
      return (Tcl_Obj*) NULL;
    }

    res = Tcl_ListObjAppendElement (interp, list, o);

    if (res != TCL_OK) {
      Tcl_DecrRefCount (list);
      return (Tcl_Obj*) NULL;
    }

    /* check next */
  }
 
  return list;
}


static int
LFromObj (interp, o, itemPtr, clientData)
Tcl_Interp* interp;
Tcl_Obj*    o;
void*       itemPtr;
ClientData  clientData;
{
  AddressAppInfo* a = (AddressAppInfo*) itemPtr;
  Tcl_Obj**       lobjs;
  Tcl_Obj**       ilobjs;
  int             length, i, res, pindex, ilength;
  char*           buf;

  res = Tcl_ListObjGetElements (interp, o, &length, &lobjs);

  if (res != TCL_OK) {
    return TCL_ERROR;
  }

  if ((length %2) != 0) {
    Tcl_AppendResult (interp,
		      "not an even number of elements in label list",
		      NULL);
    return TCL_ERROR;
  }

  memset (a->labelRenamed, '\0', 22 * sizeof (int));
  memset (a->labels,       '\0', 22 * 16);

  for (i=0; i < length; i += 2) {

    res = TclUtil_EnumEncode (interp, &labelEnum, lobjs [i], &pindex);

    if (res != TCL_OK) {
      return TCL_ERROR;
    }

    res = Tcl_ListObjGetElements (interp, lobjs [i+1], &ilength, &ilobjs);

    if (res != TCL_OK) {
      return TCL_ERROR;
    }

    if (ilength < 2) {
      Tcl_AppendResult (interp,
			"not enough information for label ",
			Tcl_GetStringFromObj (lobjs [i], NULL),
			NULL);
      return TCL_ERROR;
    }

    res = Tcl_GetIntFromObj (interp, ilobjs [0], &a->labelRenamed [pindex]);

    if (res != TCL_OK) {
      return TCL_ERROR;
    }

    buf = Tcl_GetStringFromObj (ilobjs [1], &ilength);

    if (ilength > 15) {
      ilength = 15;
    }

    memcpy (&a->labels [pindex], buf, ilength);
    a->labels [pindex][15] = '\0';
  }

  return TCL_OK;
}



static Tcl_Obj*
PToObj (interp, itemPtr, clientData)
Tcl_Interp* interp;
void*       itemPtr;
ClientData  clientData;
{
  int* pLabel = (int*) itemPtr;
  int          i;

  i = (*pLabel < 5) ? (*pLabel + 3) : (*pLabel - 5 + 19);

  return TclUtil_EnumDecode (interp, &labelEnum, i);
}



static int
PFromObj (interp, o, itemPtr, clientData)
Tcl_Interp* interp;
Tcl_Obj*    o;
void*       itemPtr;
ClientData  clientData;
{
  int* pLabel = (int*) itemPtr;
  int  res, i;

  res = TclUtil_EnumEncode (interp, &labelEnum, o, &i);

  if (res != TCL_OK) {
    return TCL_ERROR;
  }

  if ((i < 3) || ((i > 7) && (i < 19))) {
    Tcl_AppendResult (interp,
		      "not a phonelabel given",
		      NULL);
    return TCL_ERROR;    
  }

  *pLabel = (i < 8) ? (i-3) : (i-19+5);
  return TCL_OK;
}

