/*
 * category.c --
 *
 *	Definition for general packer wrappers for category information.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: category.c,v 1.1 1999/03/27 13:26:31 aku Exp $
 */

#include "category.h"


/*
 * Declarations of internal custom converter procedures.
 */

static Tcl_Obj*
RenameToObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			  void*       itemPtr,
			  ClientData  clientData));

static int
RenameFromObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			    Tcl_Obj*    o,
			    void*       itemPtr,
			    ClientData  clientData));

static Tcl_Obj*
NameToObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			void*       itemPtr,
			ClientData  clientData));

static int
NameFromObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			  Tcl_Obj*    o,
			  void*       itemPtr,
			  ClientData  clientData));


/*
 * Definitions for conversion of category application information.
 */

static TclUtil_StructItem catItems [] = {
  ITEM_CUSTOM ("renamed",      renamed,      CategoryAppInfo, RenameToObj, RenameFromObj, NULL),
  ITEM_CUSTOM ("name",         name,         CategoryAppInfo, NameToObj,   NameFromObj,   NULL),
  ITEM_CHARS  ("id",           ID,           CategoryAppInfo, MAX_CHAR),
  ITEM        ("lastUniqueId", lastUniqueID, CategoryAppInfo, TUSI_Byte),
  ITEM_END
};


TclUtil_Struct catStruct = {
  "category information", 0, NULL, catItems
};


/*
 *------------------------------------------------------*
 *
 *	RenameToObj --
 *
 *	------------------------------------------------*
 *	This procedure realizes the custom functionality
 *	for the conversion of the 'rename' information
 *	in the generic category block into an object.
 *	The result is a list containing the indices of
 *	the renamed categories.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocated memory.
 *
 *	Result:
 *		A reference to a newly created Tcl_Obj'ect.
 *
 *------------------------------------------------------*
 */

static Tcl_Obj*
RenameToObj (interp, itemPtr, clientData)
Tcl_Interp* interp;
void*       itemPtr;
ClientData  clientData; /* Dummy */
{
  int i, res;

  /* Actually it is 'renamed [MAX_CAT]' ! */

  int*     renamed = (int*) itemPtr;
  Tcl_Obj* list    = Tcl_NewListObj (0, NULL);
  Tcl_Obj* o;

  if (list == (Tcl_Obj*) NULL) {
    return (Tcl_Obj*) NULL;
  }

  for (i=0; i < MAX_CAT; i++) {
    if (renamed [i]) {
      o = Tcl_NewIntObj (i);

      if (o == (Tcl_Obj*) NULL) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      res = Tcl_ListObjAppendElement (interp, list, o);

      if (res != TCL_OK) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      /* check next */
    }
  }
 
  return list;
}

/*
 *------------------------------------------------------*
 *
 *	RenameFromObj --
 *
 *	------------------------------------------------*
 *	This procedure realizes the custom functionality
 *	for the conversion of a tcl object into the
 *	'rename' information in the generic category block.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Writes the information in the object into
 *		the given structure.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
RenameFromObj (interp, o, itemPtr, clientData)
Tcl_Interp* interp;
Tcl_Obj*    o;
void*       itemPtr;
ClientData  clientData;
{
  /* Actually its 'renamed [MAX_CAT]' ! */

  int*      renamed = (int*) itemPtr;
  Tcl_Obj** lobjs;
  int       length, i, res, k;

  res = Tcl_ListObjGetElements (interp, o, &length, &lobjs);

  if (res != TCL_OK) {
    return TCL_ERROR;
  }

  for (i=0; i < MAX_CAT; i++) {
    renamed [i] = 0;
  }

  for (i=0; i < length; i++) {
    res = Tcl_GetIntFromObj (interp, lobjs [i], &k);
    if (res != TCL_OK) {
      return TCL_ERROR;
    }

    if ((k < 0) || (MAX_CAT <= k)) {
      Tcl_AppendResult (interp, "category index out of range", NULL);
      return TCL_ERROR;
    }

    renamed [k] = 1;
  }

  return TCL_OK;
}


/*
 *------------------------------------------------------*
 *
 *	NameToObj --
 *
 *	------------------------------------------------*
 *	This procedure realizes the custom functionality
 *	for the conversion of the 'name' information
 *	in the generic category block into an object.
 *	The result is a list containing the name of the
 *	the defined categories.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocated memory.
 *
 *	Result:
 *		A reference to a newly created Tcl_Obj'ect.
 *
 *------------------------------------------------------*
 */

static Tcl_Obj*
NameToObj (interp, itemPtr, clientData)
Tcl_Interp* interp;
void*       itemPtr;
ClientData  clientData;
{
  char* name = (char*) itemPtr;

  /* name = char [MAX_CAT][MAX_CHAR]
   * MAX_CHAR = MAX_CHAR-1 characters + \0
   */

  Tcl_Obj* list    = Tcl_NewListObj (0, NULL);
  Tcl_Obj* o;
  int      i, res;

  if (list == (Tcl_Obj*) NULL) {
    return (Tcl_Obj*) NULL;
  }

  for (i=0; i < MAX_CAT; i++) {
      o = Tcl_NewStringObj (name, -1);

      if (o == (Tcl_Obj*) NULL) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      res = Tcl_ListObjAppendElement (interp, list, o);

      if (res != TCL_OK) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      /* check next */
      name += MAX_CHAR;
  }

  return list;
}

/*
 *------------------------------------------------------*
 *
 *	NameFromObj --
 *
 *	------------------------------------------------*
 *	This procedure realizes the custom functionality
 *	for the conversion of a tcl object into the 'name'
 *	information in the generic category block.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Writes the information in the object into
 *		the given structure.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
NameFromObj (interp, o, itemPtr, clientData)
Tcl_Interp* interp;
Tcl_Obj*    o;
void*       itemPtr;
ClientData  clientData;
{
  char* name = (char*) itemPtr;

  /* name = char [MAX_CAT][MAX_CHAR]
   * MAX_CHAR = MAX_CHAR-1 characters + \0
   */

  Tcl_Obj** lobjs;
  int       length, i, res, len;
  char*     buf;

  res = Tcl_ListObjGetElements (interp, o, &length, &lobjs);

  if (res != TCL_OK) {
    return res;
  }

  if (length > MAX_CAT) {
    /* Truncate superfluous list elements */
    length = MAX_CAT;
  }

  /* Prepare the playground, allow for less than MAX_CAT elements
   */

  memset (name, '\0', MAX_CAT * MAX_CHAR);

  /* And copy the strings into it, truncating as necessary.
   */

  for (i=0; i < length; i++) {
    buf = Tcl_GetStringFromObj (lobjs [i], &len);

    if (len >= MAX_CHAR) {
      len = MAX_CHAR;
    }

    strncpy (name, buf, len);
    name [MAX_CHAR-1] = '\0';

    name += MAX_CHAR;
  }

  return TCL_OK;
}

