/*
 * piTclDlp.c --
 *
 *	Implementation of the high-level DLP functionality.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: db.c,v 1.1 1999/03/27 13:26:32 aku Exp $
 */


#include <assert.h>
#include "internal.h"

/*
 * Forward declarations of internal procedures.
 */

/* Procedures implementing the various subcommands
 */

static int
DbCategory _ANSI_ARGS_ ((PiTclDbInstance* db,
			 char*            cmd,
			 Tcl_Interp*      interp,
			 int              objc, 
			 Tcl_Obj* CONST   objv []));

static int
DbSet _ANSI_ARGS_ ((PiTclDbInstance* db,
		    char*            cmd,
		    Tcl_Interp*      interp,
		    int              objc, 
		    Tcl_Obj* CONST   objv []));

static int
DbGet _ANSI_ARGS_ ((PiTclDbInstance* db,
		    char*            cmd,
		    Tcl_Interp*      interp,
		    int              objc, 
		    Tcl_Obj* CONST   objv []));

static int
DbDelete _ANSI_ARGS_ ((PiTclDbInstance* db,
		       char*            cmd,
		       Tcl_Interp*      interp,
		       int              objc, 
		       Tcl_Obj* CONST   objv []));

static int
DbForeach _ANSI_ARGS_ ((PiTclDbInstance* db,
			char*            cmd,
			Tcl_Interp*      interp,
			int              objc, 
			Tcl_Obj* CONST   objv []));


static int
DbUnpack _ANSI_ARGS_ ((PiTclDbInstance* db,
		       Tcl_Interp*      interp,
		       Tcl_Obj*         unpackCmd));

static int
DbPack _ANSI_ARGS_ ((PiTclDbInstance* db,
		     Tcl_Interp*      interp,
		     Tcl_Obj*         packCmd,
		     int              objc, 
		     Tcl_Obj* CONST   objv []));

static int
DbGetRecord _ANSI_ARGS_ ((PiTclDbInstance* db,
			  char*            cmd,
			  Tcl_Interp*      interp,
			  int              objc, 
			  Tcl_Obj* CONST   objv []));
static int
DbGetResource _ANSI_ARGS_ ((PiTclDbInstance* db,
			    char*            cmd,
			    Tcl_Interp*      interp,
			    int              objc, 
			    Tcl_Obj* CONST   objv []));
/*
static int
DbGetNextModCatRecord _ANSI_ARGS_ ((PiTclDbInstance* db,
				    char*            cmd,
				    Tcl_Interp*      interp,
				    int              objc, 
				    Tcl_Obj* CONST   objv []));
*/
static int
DbGetNextCatRecord _ANSI_ARGS_ ((PiTclDbInstance* db,
				 char*            cmd,
				 Tcl_Interp*      interp,
				 int              objc, 
				 Tcl_Obj* CONST   objv []));
static int
DbCGet _ANSI_ARGS_ ((PiTclDbInstance* db,
		     char*            cmd,
		     Tcl_Interp*      interp,
		     Tcl_Obj*         option));
static int
DbConfigure _ANSI_ARGS_ ((PiTclDbInstance* db,
			  char*            cmd,
			  Tcl_Interp*      interp,
			  int              objc, 
			  Tcl_Obj* CONST   objv []));

/*
 *------------------------------------------------------*
 *
 *	PiTclDeleteDbObjectCmd --
 *
 *	------------------------------------------------*
 *	This procedure cleans up any state left over from
 *	a deleted DB object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Resets the state of the PiTclDbInstance
 *		structure.
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

void
PiTclDeleteDbObjectCmd (cd)
ClientData cd;
{
  PiTclDbInstance* db = (PiTclDbInstance*) cd;

  if (db->sd >= 0) {
    dlp_CloseDB    (db->sd, db->handle);
    /*dlp_FindDBInfo (db->sd, 0, 0, (char*) db->name, 0, 0, &db->info);*/
    db->info.flags &= ~dlpDBFlagOpen;
  }

  db->sd     = -1;
  db->handle = -1;
  db->cmd    = (Tcl_Command) NULL;

  if (db->pack != (PiTclPackInfo*) NULL) {
    PiTclPackRelease (db->pack);
  }

  db->pack   = (PiTclPackInfo*) NULL;

  /* The structure is *not* released as it is part of
   * the database cache in the surrounding dlp object.
   */
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDlpDbObjectCmd --
 *
 *	------------------------------------------------*
 *	This procedure realizes the command of DB objects.
 *	See the manpages for details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclDlpDbObjectCmd (cd, interp, objc, objv)
ClientData     cd;
Tcl_Interp*    interp;
int            objc;
Tcl_Obj* CONST objv [];
{
  /* Convert subcommand to easy-to-use index. This converts the
   * given Tcl_Obj too, caching the conversion result. Later
   * accesses are done in constant time :-)
   */

  PiTclDbInstance* db  = (PiTclDbInstance*) cd;
  char*            cmd = ARG (0);
  int              pindex, dlpResult;

  static char *subcmd [] = {
    "category", "cleanup", "close", "configure", "cget", "delete",
    "foreach", "get", "resetIndex", "resetSync", "set", NULL
  };
  enum subcmd {
    DB_CATEGORY, DB_CLEANUP, DB_CLOSE, DB_CONFIG, DB_CGET, DB_DELETE,
    DB_FOREACH, DB_GET, DB_RESET_INDEX, DB_RESET_SYNC, DB_SET
  };	  

  DLP_CHECKARGS (objc < 2, interp, cmd, " <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Now dispatch to the appropriate subcommands
   */

  switch (pindex) {
  case DB_CLEANUP:
    DLP_CHECKARGS_S (objc > 2, interp, cmd, " ", ARG (1));
    dlpResult = dlp_CleanUpDatabase (db->sd, db->handle);
    break;

  case DB_RESET_INDEX:
    DLP_CHECKARGS_S (objc > 2, interp, cmd, " ", ARG (1));
    dlpResult = dlp_ResetDBIndex (db->sd, db->handle);
    break;

  case DB_RESET_SYNC:
    DLP_CHECKARGS_S (objc > 2, interp, cmd, " ", ARG (1));
    dlpResult = dlp_ResetSyncFlags (db->sd, db->handle);
    break;

  case DB_CLOSE:
    DLP_CHECKARGS_S (objc > 2, interp, cmd, " ", ARG (1));
    Tcl_DeleteCommandFromToken (interp, db->cmd);
    return TCL_OK;

  case DB_CATEGORY:
    return DbCategory (db, cmd, interp, objc-1, objv+1);

  case DB_DELETE:
    return DbDelete (db, cmd, interp, objc-1, objv+1);

  case DB_GET:
    return DbGet (db, cmd, interp, objc-1, objv+1);

  case DB_SET:
    return DbSet (db, cmd, interp, objc-1, objv+1);

  case DB_FOREACH:
    return DbForeach (db, cmd, interp, objc-1, objv+1);

  case DB_CONFIG:
    DLP_CHECKARGS (objc > 4, interp, cmd, " configure ?-pack ?value??");

    if (objc == 2) {
      int res;
      Tcl_Obj* list = Tcl_NewListObj (0, NULL);

      LIST_ADDSTR (list, "-pack");

      Tcl_SetObjResult (interp, list);
      return TCL_OK;
    } else if (objc == 3) {
      return DbCGet (db, cmd, interp, objv [2]);
    } else /* objc == 4 */ {
      return DbConfigure (db, cmd, interp, objc-2, objv+2);
    }

  case DB_CGET:
    DLP_CHECKARGS (objc != 3, interp, cmd, " cget -pack");

    return DbCGet (db, cmd, interp, objv [2]);


  default:
    /* impossible */
    /* We should not come to this place */
    return TCL_ERROR;
  }

  DLP_CHECK (interp, cmd, ARG(1), dlpResult);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbCategory --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'category' sub-
 *	command of db objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbCategory (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  static char *subcmd [] = {
    "delete", "move", NULL
  };
  enum subcmd {
    CAT_DELETE, CAT_MOVE
  };	  

  int pindex, dlpResult, category, fcat, tcat;

  DLP_CHECKARGS (objc < 2, interp, cmd, " category <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {
  case CAT_DELETE:
    DLP_CHECKARGS (objc != 3, interp, cmd, " category delete <cat>");

    if (Tcl_GetIntFromObj (interp, objv [2], &category) != TCL_OK) {
      return TCL_ERROR;
    }

    dlpResult = dlp_DeleteCategory (db->sd, db->handle, category);
    DLP_CHECK (interp, cmd, "category delete", dlpResult);
    break;


  case CAT_MOVE:
    DLP_CHECKARGS (objc != 4, interp, cmd, " category move <from> <to>");

    if (Tcl_GetIntFromObj (interp, objv [2], &fcat) != TCL_OK) {
      return TCL_ERROR;
    }

    if (Tcl_GetIntFromObj (interp, objv [3], &tcat) != TCL_OK) {
      return TCL_ERROR;
    }

    dlpResult = dlp_MoveCategory (db->sd, db->handle, fcat, tcat);
    DLP_CHECK (interp, cmd, "category move", dlpResult);
    break;


  default:
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbSet --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'set' sub-
 *	command of db objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbSet (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  /* (db) set appinfo  <data>
   * (db) set sortinfo <data>
   * (db) set record <id> <data>
   *                     ~~~~~~ = {attr cat <data>}
   *                       or = {type <data>}
   */

  static char *subcmd [] = {
    "appinfo", "record", "resource", "sortinfo", NULL
  };
  enum subcmd {
    SET_APPINFO, SET_RECORD, SET_RESOURCE, SET_SORTINFO
  };	  

  int pindex, dlpResult;
  Tcl_Obj*   packCmd = (Tcl_Obj*) NULL;

  DLP_CHECKARGS (objc < 2, interp, cmd, " set <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Prepare for gathering of data, gather
   * and pack it, then write it to the pilot.
   */

  switch (pindex) {
  case SET_APPINFO:
    DLP_CHECKARGS (objc > 3, interp, cmd, " set appinfo data");
    db->isRecord = 0;

    if (db->pack != (PiTclPackInfo*) NULL) {
      packCmd = db->pack->pAppInfoCmd;
    }
    break;


  case SET_SORTINFO:
    DLP_CHECKARGS (objc > 3, interp, cmd, " set sortinfo data");
    db->isRecord = 0;

    if (db->pack != (PiTclPackInfo*) NULL) {
      packCmd = db->pack->pSortInfoCmd;
    }
    break;


  case SET_RECORD:
    DLP_CHECKARGS (objc > 4, interp, cmd, " set record id {attr cat data}");
    if (db->info.flags & dlpDBFlagResource) {
      Tcl_AppendResult (interp, "'set record' cannot operate ",
			"on a resource database",
			(char*) NULL);
      return TCL_ERROR;
    }
    db->isRecord = 1;

    if (db->pack != (PiTclPackInfo*) NULL) {
      packCmd = db->pack->pRecordCmd;
    }
    break;


  case SET_RESOURCE:
    DLP_CHECKARGS (objc > 4, interp, cmd, " set resource id {type data}");
    if (! (db->info.flags & dlpDBFlagResource)) {
      Tcl_AppendResult (interp, "'set resource' cannot ",
			"operate on a record database",
			(char*) NULL);
      return TCL_ERROR;
    }
    db->isRecord = 2;

    if (db->pack != (PiTclPackInfo*) NULL) {
      packCmd = db->pack->pRecordCmd;
    }
    break;

  default:
    /* Impossible */
    return TCL_ERROR;
  }

  if (DbPack (db, interp, packCmd, objc-2, objv+2) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {
  case SET_APPINFO:
    dlpResult = dlp_WriteAppBlock (db->sd, db->handle, db->buf, db->buflen);

    DLP_CHECK (interp, cmd, "set appinfo", dlpResult);
    return TCL_OK;


  case SET_SORTINFO:
    dlpResult = dlp_WriteSortBlock (db->sd, db->handle, db->buf, db->buflen);

    DLP_CHECK (interp, cmd, "set sortinfo", dlpResult);
    return TCL_OK;


  case SET_RECORD:
    dlpResult = dlp_WriteRecord (db->sd, db->handle, db->attr,
				 db->id, db->category,
				 db->buf, db->buflen, &db->id);
    DLP_CHECK (interp, cmd, "set record", dlpResult);

    /* Return id as given by the pilot */
    Tcl_SetObjResult (interp, Tcl_NewIntObj (db->id));
    return TCL_OK;


  case SET_RESOURCE:
    dlpResult = dlp_WriteResource (db->sd, db->handle, db->resType,
				   db->id, db->buf, db->buflen);
    DLP_CHECK (interp, cmd, "set resource", dlpResult);

    /* Return chosen id */
    Tcl_SetObjResult (interp, Tcl_NewIntObj (db->id));
    return TCL_OK;


  default:
    /* impossible to reach */
    return TCL_ERROR;
  }

  /* impossible to reach */
  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	DbGet --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'get' subcommand of
 *	db objects. See the manpages for details on what
 *	it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbGet (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  static char *subcmd [] = {
    "appinfo", "info", "nextModified", "nextModifiedInCategory",
    "nextInCategory", "recnum", "record", "recordlist",
    "resource", "sortinfo", NULL
  };
  enum subcmd {
    GET_APPINFO, GET_INFO, GET_NEXTMOD, GET_NEXTMODCAT, GET_NEXTCAT,
    GET_RECNUM, GET_RECORD, GET_RECLIST, GET_RESOURCE, GET_SORTINFO
  };	  

  int pindex, dlpResult, recnum, size;
  Tcl_Obj*   unpackCmd = (Tcl_Obj*) NULL;


  DLP_CHECKARGS (objc < 2, interp, cmd, " get <subcommand>...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {
  case GET_INFO:
    DLP_CHECKARGS (objc > 2, interp, cmd, " get info");

    return PiTclDbObjectToList (interp, db);


  case GET_RECNUM:
    DLP_CHECKARGS (objc > 2, interp, cmd, " get recnum");

    dlpResult = dlp_ReadOpenDBInfo (db->sd, db->handle, &recnum);

    DLP_CHECK (interp, cmd, "get recnum", dlpResult);

    Tcl_SetObjResult (interp, Tcl_NewIntObj (recnum));
    return TCL_OK;


  case GET_RECLIST:
    {
      Tcl_Obj*    list  = Tcl_NewListObj (0, NULL);
      int         start = 0;
      int         sort  = 0;
      int         count, i, res;
      recordid_t* id    = (recordid_t*) db->buf;

      DLP_CHECKARGS (objc > 3, interp, cmd, " get recordlist ?-sorted?");

      if (objc == 3) {
	if (strcmp ("-sorted", Tcl_GetStringFromObj (objv [2], NULL)) != 0) {
	  Tcl_AppendResult (interp, "unknown option ",
			    ARG (2), NULL);
	  return TCL_ERROR;
	} else {
	  sort = 1;
	}
      }

#define MAXID (MAXBUF / sizeof (recordid_t))

      for (;;) {
	dlpResult = dlp_ReadRecordIDList (db->sd, db->handle, sort, start,
					  MAXID, id, &count);

	if (dlpResult < 0) {
	  Tcl_DecrRefCount (list);
	}
	DLP_CHECK (interp, cmd, "get recordlist", dlpResult);

	for (i=0; i < count; i++) {
	  LIST_ADDINT (list, id [i]);
	}

	if (count == MAXID) {
	  start = count;
	} else {
	  break;
	}
      }

      Tcl_SetObjResult (interp, list);
#undef MAXID
    }
  return TCL_OK;


  case GET_APPINFO:
    DLP_CHECKARGS (objc > 2, interp, cmd, " get appinfo");

    dlpResult = dlp_ReadAppBlock (db->sd, db->handle, 0, db->buf, MAXBUF);

    DLP_CHECK (interp, cmd, "get appinfo", dlpResult);

    db->isRecord = 0;
    db->buflen   = dlpResult;

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uAppInfoCmd;
    }
    break;


  case GET_SORTINFO:

    DLP_CHECKARGS (objc > 2, interp, cmd, " get sortinfo");

    dlpResult = dlp_ReadSortBlock (db->sd, db->handle, 0, db->buf, MAXBUF);

    DLP_CHECK (interp, cmd, "get sortinfo", dlpResult);

    db->isRecord = 0;
    db->buflen   = dlpResult;

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uSortInfoCmd;
    }
    break;


  case GET_RECORD:
    if (DbGetRecord (db, cmd, interp, objc-1, objv+1) != TCL_OK) {
      return TCL_ERROR;
    }

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uRecordCmd;
    }
    break;


  case GET_RESOURCE:
    if (DbGetResource (db, cmd, interp, objc-1, objv+1) != TCL_OK) {
      return TCL_ERROR;
    }

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uRecordCmd;
    }
    break;


  case GET_NEXTMOD:
    if (db->info.flags & dlpDBFlagResource) {
      Tcl_AppendResult (interp, cmd, "'get nextModified' not ",
			"available for resource databases",
			(char*) NULL);
      return TCL_ERROR;
    }

    DLP_CHECKARGS (objc > 3, interp, cmd, " get nextModified ?cat?");

    if (objc == 2) {
      dlpResult = dlp_ReadNextModifiedRec (db->sd, db->handle, db->buf,
					   &db->id, &pindex, &size,
					   &db->attr, &db->category);
    } else /* > 2, == 3 */ {
      int dummy;

      if (Tcl_GetIntFromObj (interp, objv [2], &db->category) != TCL_OK) {
	return TCL_ERROR;
      }
      if (db->info.flags & dlpDBFlagResource) {
	Tcl_AppendResult (interp, cmd, "'get nextModified cat' not ",
			  "available for resource databases",
			  (char*) NULL);
	return TCL_ERROR;
      }

      dlpResult = dlp_ReadNextModifiedRecInCategory (db->sd, db->handle,
						     db->category, db->buf,
						     &db->id, &dummy, &dummy,
						     &db->attr);
    }

    DLP_CHECK (interp, cmd, "get nextModified", dlpResult);

    db->isRecord = 1;
    db->buflen   = dlpResult;

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uRecordCmd;
    }
    break;

    /*
  case GET_NEXTMODCAT:
    if (db->info.flags & dlpDBFlagResource) {
      Tcl_AppendResult (interp, cmd, "'get nextModifiedInCategory' not ",
			"available for resource databases",
			(char*) NULL);
      return TCL_ERROR;
    }

    if (DbGetNextModCatRecord (db, cmd, interp, objc-1, objv+1) != TCL_OK) {
      return TCL_ERROR;
    }

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uRecordCmd;
    }
    break;
    */

  case GET_NEXTCAT:
    if (db->info.flags & dlpDBFlagResource) {
      Tcl_AppendResult (interp, cmd, "'get nextInCategory' not ",
			"available for resource databases",
			(char*) NULL);
      return TCL_ERROR;
    }

    if (DbGetNextCatRecord (db, cmd, interp, objc-1, objv+1) != TCL_OK) {
      return TCL_ERROR;
    }

    if (db->pack != (PiTclPackInfo*) NULL) {
      unpackCmd = db->pack->uRecordCmd;
    }
    break;

  default:
    return TCL_ERROR;
  }

  /* Convert (unpack) the information in the transport buffer
   */

  return DbUnpack (db, interp, unpackCmd);
}

/*
 *------------------------------------------------------*
 *
 *	DbDelete --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'delete' subcommand
 *	of db objects. See the manpages for details on
 *	what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbDelete (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  /*
   * (db) delete record|resource <id>
   * (db) delete record|resource all
   */

  int dlpResult, pindex, all, res, idpos;
  /* -W- hackery, 'int' is not necessarily 'long' (recordid_t) */

  static char* subcmd [] = {
    "record", "resource", NULL
  };
  enum subcmd {
    DEL_REC, DEL_RES
  };

  DLP_CHECKARGS (objc < 2, interp, cmd, " delete subcmd ...");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  res   = pindex == DEL_RES ? 1 : 0;
  all   = 0;
  pindex = 0;

  if (!res) {
    if (db->info.flags & dlpDBFlagResource) {
      Tcl_AppendResult (interp, cmd, "'delete record' cannot ",
			"operate on a resource database",
			(char*) NULL);
      return TCL_ERROR;
    }

    DLP_CHECKARGS (objc > 3, interp, cmd, " delete record id|'all'");
    idpos = 2;
  } else {
    if (! (db->info.flags & dlpDBFlagResource)) {
      Tcl_AppendResult (interp, cmd, "'delete resource' cannot ",
			"operate on a record database",
			(char*) NULL);
      return TCL_ERROR;
    }

    DLP_CHECKARGS (objc > 4, interp, cmd, " delete resource type id|'all'");
    idpos = 3;
  }

  if (strcmp ("all", Tcl_GetStringFromObj (objv [idpos], NULL)) == 0) {
    all = 1;
  } else if (Tcl_GetIntFromObj (interp, objv [idpos], &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  if (res) {
    long int type = PiTclStringToLong (objv [idpos-1]);

    dlpResult = dlp_DeleteResource (db->sd, db->handle, all, type, pindex);
    DLP_CHECK (interp, cmd, "delete resource", dlpResult);
  } else {
    dlpResult = dlp_DeleteRecord (db->sd, db->handle, all, pindex);
    DLP_CHECK (interp, cmd, "delete record", dlpResult);
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbForeach --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'foreach' subcommand
 *	of db objects. See the manpages for details on
 *	what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbForeach (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  /* -W- iterators */
  Tcl_AppendResult (interp,
		    "foreach not implemented yet",
		    (char*) NULL);

  return TCL_ERROR;
}

/*
 *------------------------------------------------------*
 *
 *	DbPack --
 *
 *	------------------------------------------------*
 *	This procedure converts the incoming tcl objects
 *	into binary data ready for shipping to the pilot.
 *	This effort utilizes the packer associated to the
 *	db object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Changes the transport associated state
 *		of the db object.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbPack (db, interp, packCmd, objc, objv)
PiTclDbInstance* db;
Tcl_Interp*      interp;
Tcl_Obj*         packCmd;
int              objc;
Tcl_Obj* CONST   objv [];
{
  /* objc, objv already checked
   */

  Tcl_Obj*  data;    /* Reference to main information block   */
  Tcl_Obj*  result;  /* Reference to result of packing 'data' */
  Tcl_Obj** lobjs;
  char*     buf;
  int       length, pindex;

  switch (db->isRecord) {
  case 0:
    /* appinfo / sortinfo <data>
     */

    data = objv [0];
    break;


  case 1:
    /* data record
     * id {category, attr-list, record-data}
     * 2 arguments!
     */

    /* Extract 'id' first */
    if (Tcl_GetIntFromObj (interp, objv [0],
			   &pindex) != TCL_OK) {
      return TCL_ERROR;
    }

    db->id = pindex;
    /* -W- hackery, int and long are not necessarily equal */

    /* Now get at the {category, attr-list, record-data} list */
    if (Tcl_ListObjGetElements (interp, objv [1],
				&length, &lobjs) != TCL_OK) {
      return TCL_ERROR;
    }

    /* Extract the category information from this list */
    if (Tcl_GetIntFromObj (interp, lobjs [0],
			   &db->category) != TCL_OK) {
      return TCL_ERROR;
    }

    /* Remember the main information block */
    data = lobjs [2];

    if (PiTclEncodeRecAttr (interp, lobjs [1], &db->attr) != TCL_OK) {
      return TCL_ERROR;
    }
    break;


  case 2:
    /* resource record
     * {resourceType, record-data}
     */

    /* Extract 'id' first */
    if (Tcl_GetIntFromObj (interp, objv [0],
			   &pindex) != TCL_OK) {
      return TCL_ERROR;
    }

    db->id = pindex;
    /* -W- hackery, int and long are not necessarily equal */

    /* Get at the {resourceType, record-data} list */
    if (Tcl_ListObjGetElements (interp, objv [1],
				&length, &lobjs) != TCL_OK) {
      return TCL_ERROR;
    }

    /* Remember the main information block */
    data = lobjs [1];

    /* Decode the resource type */
    db->resType = PiTclStringToLong (lobjs [0]);
    break;

  default:
    return TCL_ERROR;
  }

  /* Construct the binary representation of the main information
   * block and placee it into the transport buffers.
   */

  if (packCmd == (Tcl_Obj*) NULL) {
    /* No packer available, assume that the data is already
     * in the correct binary format.
     */

    result = data;

  } else {
    int res;

    packCmd = Tcl_DuplicateObj (packCmd);

    Tcl_IncrRefCount (packCmd);
    LIST_ADDOBJ      (packCmd, data);

    res = Tcl_EvalObj (interp, packCmd);
    Tcl_DecrRefCount (packCmd);

    if (res != TCL_OK) {
      return res;
    }

    result = Tcl_GetObjResult (interp);
  }

  buf = Tcl_GetStringFromObj (result, &length);

  if (length > MAXBUF) {
    length = MAXBUF;
  }

  memcpy (db->buf, buf, length);
  db->buflen = length;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbUnpack --
 *
 *	------------------------------------------------*
 *	This procedure converts the binary data sitting
 *	in the transport buffer into a readable list, by
 *	using the packer associated to the db object.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Sets the interpreter result to the
 *		unpacked data from the pilot.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbUnpack (db, interp, unpackCmd)
PiTclDbInstance* db;
Tcl_Interp*      interp;
Tcl_Obj*         unpackCmd;
{
  int res;

  if (unpackCmd != (Tcl_Obj*) NULL) {
    unpackCmd = Tcl_DuplicateObj (unpackCmd);

    Tcl_IncrRefCount (unpackCmd);

    res = Tcl_ListObjAppendElement (interp, unpackCmd,
				    Tcl_NewStringObj (db->buf, db->buflen));
    if (res != TCL_OK) {
      Tcl_DecrRefCount (unpackCmd);
      return TCL_ERROR;
    }

    res = Tcl_EvalObj (interp, unpackCmd);
    Tcl_DecrRefCount (unpackCmd);

    if (res != TCL_OK) {
      return res;
    }
  } else {
    /* No unpacker available, resort to a binary result.
     */

    Tcl_SetObjResult (interp, Tcl_NewStringObj (db->buf, db->buflen));
    res = TCL_OK;
  }

  /* The result of the eval is the current interp
   * result, as it should be.
   */

  if (db->isRecord == 1) {
    /* Not quite true for a record.
     * Here we have to blow the result up into a list containing:
     * {id, category, attr-list, record-data}
     */

    Tcl_Obj* data = Tcl_GetObjResult (interp);
    Tcl_Obj* attr = PiTclDecodeRecAttr (interp, db->attr);
    Tcl_Obj* list;

    if (attr == (Tcl_Obj*) NULL) {
      return TCL_ERROR;
    }

    list = Tcl_NewListObj (0, NULL);

    LIST_ADDINT (list, db->id);
    LIST_ADDINT (list, db->category);
    LIST_ADDOBJ (list, attr);
    LIST_ADDOBJ (list, data);

    Tcl_SetObjResult (interp, list);
  } else if (db->isRecord == 2) {
    /* Not quite true for a resource block.
     * Here we have to blow the result up into a list containing:
     * {resourceType, record-data}
     */

    Tcl_Obj* data = Tcl_GetObjResult (interp);
    Tcl_Obj* list = Tcl_NewListObj (0, NULL);

    LIST_ADDOBJ (list, PiTclLongToString (db->resType));
    LIST_ADDOBJ (list, data);

    Tcl_SetObjResult (interp, list);
  }

  return res;
}

/*
 *------------------------------------------------------*
 *
 *	DbGetResource --
 *
 *	------------------------------------------------*
 *	This procedure realizes the retrieval part of the
 *	'db get resource' subcommand of db objects. See
 *	the manpages for details on what it does. The caller
 *	will invoke 'DbUnpack' to finish the processing.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbGetResource (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  int         dlpResult = 0;
  int         idx;
  int         dummy;

  /* syntax
   * (db get) resource idx
   */

  if (! (db->info.flags & dlpDBFlagResource)) {
      Tcl_AppendResult (interp, cmd, " get resource not",
			" available for record databases",
			(char*) NULL);
      return TCL_ERROR;
  }

  DLP_CHECKARGS (objc > 2, interp, cmd, " get resource idx");

  if (Tcl_GetIntFromObj (interp, objv [1], &idx) != TCL_OK) {
    return TCL_ERROR;
  }

  dlpResult = dlp_ReadResourceByIndex (db->sd, db->handle, idx,
				       db->buf, &db->resType,
				       &dummy, &dummy);

  DLP_CHECK (interp, cmd, "get resource", dlpResult);

  db->isRecord = 2;
  db->buflen   = dlpResult;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbGetRecord --
 *
 *	------------------------------------------------*
 *	This procedure realizes the retrieval part of the
 *	'db get record' subcommand of db objects. See the
 *	manpages for details on what it does. The caller
 *	will invoke 'DbUnpack' to finish the processing.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbGetRecord (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  int         dlpResult = 0;
  int         byId      = 0;
  int         id;
  int         dummy;
  Tcl_Obj*    idObj;

  /* syntax
   * (db get) record ?-byId? <id|idx>
   */

  if (db->info.flags & dlpDBFlagResource) {
    Tcl_AppendResult (interp, cmd, " get record not ",
		      "available for resource databases",
		      (char*) NULL);
    return TCL_ERROR;
  }

  DLP_CHECKARGS (objc > 3, interp, cmd, " get record ?-byId? <id|idx>");

  if (strcmp ("-byId", ARG(1)) == 0) {
    idObj = objv [2];
    byId  = 1;
  } else {
    DLP_CHECKARGS (objc > 2, interp, cmd, " get record idx");

    idObj = objv [1];
  }

  if (Tcl_GetIntFromObj (interp, idObj, &id) != TCL_OK) {
    return TCL_ERROR;
  }

  if (byId) {
      dlpResult = dlp_ReadRecordById (db->sd, db->handle, id,
				      db->buf, &dummy, &dummy,
				      &db->attr, &db->category);
      db->id = id;
  } else {
    dlpResult = dlp_ReadRecordByIndex (db->sd, db->handle, id,
				       db->buf, &db->id, &dummy,
				       &db->attr, &db->category);
  }

  DLP_CHECK (interp, cmd, "get record", dlpResult);

  db->isRecord = 1;
  db->buflen   = dlpResult;

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbGetNextModCatRecord --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'system card' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 * /

static int
DbGetNextModCatRecord (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  int dlpResult = 0;
  int dummy;

  DLP_CHECKARGS (objc != 3, interp, cmd, " get nextModifiedInCategory <cat>");

  if (Tcl_GetIntFromObj (interp, objv [2], &db->category)) {
    return TCL_ERROR;
  }

  dlpResult = dlp_ReadNextModifiedRecInCategory (db->sd, db->handle,
						 db->category, db->buf,
						 &db->id, &dummy, &dummy,
						 &db->attr);

  DLP_CHECK (interp, cmd, "get nextModifiedInCategory", dlpResult);

  db->isRecord = 1;
  db->buflen   = dlpResult;
    
  return TCL_OK;
}*/

/*
 *------------------------------------------------------*
 *
 *	DbGetNextCatRecord --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'system card' sub-
 *	command of dlp objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbGetNextCatRecord (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  int dlpResult = 0;
  int dummy;

  DLP_CHECKARGS (objc != 3, interp, cmd, " get nextInCategory <cat>");

  if (Tcl_GetIntFromObj (interp, objv [2], &db->category)) {
    return TCL_ERROR;
  }

  dlpResult = dlp_ReadNextRecInCategory (db->sd, db->handle, db->category,
					 db->buf, &db->id, &dummy, &dummy,
					 &db->attr);

  DLP_CHECK (interp, cmd, "get nextInCategory", dlpResult);

  db->isRecord = 1;
  db->buflen   = dlpResult;
    
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbCGet --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'cget' subcommand of
 *	db objects. See the manpages for details on what
 *	it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbCGet (db, cmd, interp, option)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
Tcl_Obj*         option;
{
  static char *options [] = {
    "-pack", NULL
  };
  enum options {
    DB_OPACK
  };

  int pindex;

  if (Tcl_GetIndexFromObj(interp, option, options, "option", 0,
			  &pindex) != TCL_OK) {
    return TCL_ERROR;
  }

  switch (pindex) {
  case DB_OPACK:
    if (db->pack == (PiTclPackInfo*) NULL) {
      Tcl_SetObjResult (interp, Tcl_NewStringObj ("", -1));
    } else {
      Tcl_SetObjResult (interp, Tcl_NewStringObj (db->pack->name, -1));
    }
    break;

  default:
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	DbConfigure --
 *
 *	------------------------------------------------*
 *	This procedure realizes the 'configure' sub-
 *	command of db objects. See the manpages for
 *	details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
DbConfigure (db, cmd, interp, objc, objv)
PiTclDbInstance* db;
char*            cmd;
Tcl_Interp*      interp;
int              objc;
Tcl_Obj* CONST   objv [];
{
  static char *options [] = {
    "-pack", NULL
  };
  enum options {
    DB_OPACK
  };

  int                   pindex;
  PiTclPackInfo* p;

  DLP_CHECKARGS ((objc % 2) == 1, interp, cmd, " an even number");


  for (; objc > 0; objc -= 2, objv += 2) {
    if (Tcl_GetIndexFromObj(interp, objv [0], options, "option", 0,
			    &pindex) != TCL_OK) {
      return TCL_ERROR;
    }

    switch (pindex) {
    case DB_OPACK:
      if (PiTclPackFind (&db->dlp->base.info->pack, interp, objv [1], &p) != TCL_OK) {
	return TCL_ERROR;
      } else if (p == (PiTclPackInfo*) NULL) {
	Tcl_AppendResult (interp,
			  "packer not found",
			  (char*) NULL);
	return TCL_ERROR;
      }

      if (db->pack != (PiTclPackInfo*) NULL) {
	PiTclPackRelease (db->pack);
      }

      db->pack = p;
      PiTclPackPreserve (db->pack);
      break;

    default:
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}
