/*
 * piTclInit.c --
 *
 *	Initalization of piTcl, the tcl binding to pilot-link.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: init.c,v 1.1 1999/03/27 13:26:33 aku Exp $
 */

#include "internal.h"


/*
 *------------------------------------------------------*
 *
 *	Pitcl_Init --
 *
 *	------------------------------------------------*
 *	Standard procedure required by 'load'. 
 *	Initializes this extension.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		As of 'Tcl_CreateCommand'.
 *
 *	Result:
 *		A standard Tcl error code.
 *
 *------------------------------------------------------*
 */

EXTERN EXPORT (int,Pitcl_Init) (interp)
Tcl_Interp* interp;
{
  /* Register the commands in this binding.
   */

  PiTclInfo* info;

#if GT81
  if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
    return TCL_ERROR;
  }
#endif

  info = (PiTclInfo*) Tcl_Alloc (sizeof (PiTclInfo));
  info->refCount = 1; 

  Tcl_InitHashTable (&info->objects, TCL_STRING_KEYS);
  PiTclPackInit     (&info->pack);
  PiTclMapInit      (&info->map);

  Tcl_CreateObjCommand (interp, "piDlp",
			&PiTclDlpCmd,
			(ClientData) info,
			&PiTclDeleteCommand);

  if (PiTclInitMemo (interp) != TCL_OK) {
    return TCL_ERROR;
  }

  if (PiTclInitTodo (interp) != TCL_OK) {
    return TCL_ERROR;
  }

  if (PiTclInitAppointment (interp) != TCL_OK) {
    return TCL_ERROR;
  }

  if (PiTclInitAddress (interp) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Register us as newly available package */
  Tcl_PkgProvide (interp, "Pitcl", PITCL_VERSION);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	Pitcl_SafeInit --
 *
 *	------------------------------------------------*
 *	Standard procedure required by 'load'. 
 *	Initializes this extension for a safe interpreter.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		As of 'Pitcl_Init'
 *
 *	Result:
 *		A standard Tcl error code.
 *
 *------------------------------------------------------*
 */

EXTERN EXPORT (int,Pitcl_SafeInit) (interp)
Tcl_Interp* interp;
{
  return Pitcl_Init (interp);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDeleteCommand --
 *
 *	------------------------------------------------*
 *	Called whenever one of the commands of this
 *	binding is deleted. Decrements the counter
 *	holding the number of commands using the client
 *	data structure and releases it if it is not used
 *	anymore.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		May release memory.
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

void
PiTclDeleteCommand (cd)
ClientData cd;
{
  PiTclInfo* info = (PiTclInfo*) cd;

  info->refCount --;

  if (info->refCount <= 0) {

    Tcl_DeleteHashTable (&info->objects);
    PiTclPackFree       (&info->pack);
    PiTclMapDelete      (&info->map);

    /* check wether the entries are deleted by this or not */
    Tcl_Free ((char*) info);
  }
}

/*
 *------------------------------------------------------*
 *
 *	PiTclDeviceIsInUse --
 *
 *	------------------------------------------------*
 *	Checks the global datastructures for entry with
 *	the given device as key.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		1 if the given device is already used by
 *		an object, 0 else.
 *
 *------------------------------------------------------*
 */

EXTERN int
PiTclDeviceIsInUse (info, deviceName)
PiTclInfo* info;
char*      deviceName;
{
  Tcl_HashEntry* hPtr;

  hPtr = Tcl_FindHashEntry (&info->objects, deviceName);

  if (hPtr == (Tcl_HashEntry*) NULL) {
    return 0;
  } else {
    return 1;
  }
}
