/*
 * pack.c --
 *
 *	Implementation of packer management.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: pack.c,v 1.1 1999/03/27 13:26:34 aku Exp $
 */

#include "pack.h"
#include "macros.h"

/* Definitions used by the internal option handlers.
 */

static char *options [] = {
  "-appinfoPack",	"-appinfoUnpack",
  "-recordPack",	"-recordUnpack",
  "-sortinfoPack",	"-sortinfoUnpack",
  NULL
};

enum options {
  O_AI_P, O_AI_U,
  O_RE_P, O_RE_U,
  O_SI_P, O_SI_U
};

/*
 *------------------------------------------------------*
 *
 *      PiTclPackInit --
 *
 *	------------------------------------------------*
 *	This procedure initializes the table of packers.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the given structure.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackInit (p)
PiTclPackers* p;
{
  Tcl_InitHashTable (&p->cvt, TCL_STRING_KEYS);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackFree --
 *
 *	------------------------------------------------*
 *	This procedure frees all resources of the given
 *	packer table.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackFree (p)
PiTclPackers* p;
{
  Tcl_HashSearch hs;
  Tcl_HashEntry* h;

  for (h = Tcl_FirstHashEntry (&p->cvt, &hs);
       h != (Tcl_HashEntry*) NULL;
       h = Tcl_NextHashEntry (&hs)) {
    
    PiTclPackRelease ((PiTclPackInfo*) Tcl_GetHashValue (h));
  }

  Tcl_DeleteHashTable (&p->cvt);
  return TCL_OK;
}
/*
 *------------------------------------------------------*
 *
 *	PiTclPackCreate --
 *
 *	------------------------------------------------*
 *	This procedure creates an empty packer information
 *	block. Its reference count is initialized to zero.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates and initializes memory.
 *
 *	Result:
 *		A pointer to the allocated block, or NULL
 *
 *------------------------------------------------------*
 */

PiTclPackInfo*
PiTclPackCreate (name)
char* name;
{
  PiTclPackInfo* p;

  p = (PiTclPackInfo*) Tcl_Alloc (sizeof (PiTclPackInfo));

  p->name         = DUP (name);
  p->refCount     = 0;
  p->hPtr         = (Tcl_HashEntry*) NULL;
  p->pRecordCmd   = (Tcl_Obj*) NULL;
  p->uRecordCmd   = (Tcl_Obj*) NULL;
  p->pAppInfoCmd  = (Tcl_Obj*) NULL;
  p->uAppInfoCmd  = (Tcl_Obj*) NULL;
  p->pSortInfoCmd = (Tcl_Obj*) NULL;
  p->uSortInfoCmd = (Tcl_Obj*) NULL;

  return p;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackRelease --
 *
 *	------------------------------------------------*
 *	This procedure releases a packer information
 *	block (if its reference count drops below 1).
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		May release memory allocated by
 *		'PiTclPackCreate' and 'PiTclPackConfigure'.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackRelease (p)
PiTclPackInfo* p;
{
  p->refCount --;

  if (p->refCount > 0) {
    return TCL_OK;
  }

  /* refCount <= 0 */

  if (p->pRecordCmd   != (Tcl_Obj*) NULL) {
    Tcl_DecrRefCount (p->pRecordCmd);
  }
  if (p->uRecordCmd   != (Tcl_Obj*) NULL) {
    Tcl_DecrRefCount (p->uRecordCmd);
  }
  if (p->pAppInfoCmd  != (Tcl_Obj*) NULL) {
    Tcl_DecrRefCount (p->pAppInfoCmd);
  }
  if (p->uAppInfoCmd  != (Tcl_Obj*) NULL) {
    Tcl_DecrRefCount (p->uAppInfoCmd);
  }
  if (p->pSortInfoCmd != (Tcl_Obj*) NULL) {
    Tcl_DecrRefCount (p->pSortInfoCmd);
  }
  if (p->uSortInfoCmd != (Tcl_Obj*) NULL) {
    Tcl_DecrRefCount (p->uSortInfoCmd);
  }

  Tcl_Free (p->name);
  Tcl_Free ((char*) p);

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackPreserve --
 *
 *	------------------------------------------------*
 *	This procedure is called to register another
 *	reference to a packer information block. Use
 *	'PiTclPackRelease' to negate this call.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Increments the refCount
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

void
PiTclPackPreserve (p)
PiTclPackInfo* p;
{
  p->refCount ++;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackConfigure --
 *
 *	------------------------------------------------*
 *	This procedure changes the contents of a packer
 *	information block according to the specified
 *	option/value pairs.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the packer information block.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackConfigure (p, interp, cmd, objc, objv)
PiTclPackInfo* p;
Tcl_Interp*      interp;
char*               cmd;
int                objc;
Tcl_Obj* CONST     objv [];
{
  int      index;
  Tcl_Obj* old;
  Tcl_Obj* new;

  DLP_CHECKARGS ((objc % 2) == 1, interp, "", "an even number");

  for (; objc > 0; objc -= 2, objv += 2) {
    if (Tcl_GetIndexFromObj(interp, objv [0], options, "option", 0,
			    &index) != TCL_OK) {
      return TCL_ERROR;
    }

    new = objv [1];
    Tcl_IncrRefCount (new);

    switch (index) {
    case O_AI_P:
      old = p->pAppInfoCmd;
      p->pAppInfoCmd = new;
      break;

    case O_AI_U:
      old = p->uAppInfoCmd;
      p->uAppInfoCmd = new;
      break;

    case O_RE_P:
      old = p->pRecordCmd;
      p->pRecordCmd = new;
      break;

    case O_RE_U:
      old = p->uRecordCmd;
      p->uRecordCmd = new;
      break;

    case O_SI_P:
      old = p->pSortInfoCmd;
      p->pSortInfoCmd = new;
      break;

    case O_SI_U:
      old = p->uSortInfoCmd;
      p->uSortInfoCmd = new;
      break;


    default:
      return TCL_ERROR;
    }

    if (old != (Tcl_Obj*) NULL) {
      Tcl_DecrRefCount (old);
    }
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackCGet --
 *
 *	------------------------------------------------*
 *	This procedure retrieves the information
 *	associated to an option
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the packer information block.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
PiTclPackCGet (p, interp, cmd, option)
PiTclPackInfo* p;
Tcl_Interp*      interp;
char*               cmd;
Tcl_Obj*         option;
{
  int      index;
  Tcl_Obj* o;

  if (Tcl_GetIndexFromObj(interp, option, options, "option", 0,
			  &index) != TCL_OK) {
    return (Tcl_Obj*) NULL;
  }

  switch (index) {
  case O_AI_P:    o = p->pAppInfoCmd;     break;
  case O_AI_U:    o = p->uAppInfoCmd;     break;
  case O_RE_P:    o = p->pRecordCmd;      break;
  case O_RE_U:    o = p->uRecordCmd;      break;
  case O_SI_P:    o = p->pSortInfoCmd;    break;
  case O_SI_U:    o = p->uSortInfoCmd;    break;

  default:
    /* impossible */
    return (Tcl_Obj*) NULL;
  }

  if (o == (Tcl_Obj*) NULL) {
    /* No command associated to the packer,
     * return an empty string
     */

    o = Tcl_NewStringObj ("", -1);
  }

  return o;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackFind --
 *
 *	------------------------------------------------*
 *	Searches the packer table by name.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Sets a reference to the found packer
 *		information into 'res', or NULL.
 *
 *	Result:
 *		A standard TCL result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackFind (p, interp, name, res)
PiTclPackers*          p;
Tcl_Interp*            interp;
Tcl_Obj*               name;
PiTclPackInfo** res;
{
  int   len;
  char* str = Tcl_GetStringFromObj (name, &len);

  if (len == 0) {
    /* Empty string, user explicitly requests that no packer
     * is used.
     */

    *res = (PiTclPackInfo*) NULL;
  } else {
    Tcl_HashEntry* hPtr;
    hPtr = Tcl_FindHashEntry (&p->cvt, str);

    if (hPtr == (Tcl_HashEntry*) NULL) {
      Tcl_AppendResult (interp,
			"packer not found",
			(char*) NULL);
      *res = (PiTclPackInfo*) NULL;
      return TCL_ERROR;
    }

    *res = (PiTclPackInfo*) Tcl_GetHashValue (hPtr);
  }

  return TCL_OK;
}
/*
 *------------------------------------------------------*
 *
 *	PiTclPackList --
 *
 *	------------------------------------------------*
 *	Lists all entries in the packer table.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Sets the result of the specified
 *		interpreter to the constructed list.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackList (interp, p)
Tcl_Interp*    interp;
PiTclPackers*  p;
{
  Tcl_Obj*              list;
  Tcl_HashSearch        hs;
  Tcl_HashEntry*        h;
  int                   res;
  PiTclPackInfo* pa;


  list = Tcl_NewListObj (0, NULL);

  for (h = Tcl_FirstHashEntry (&p->cvt, &hs);
       h != (Tcl_HashEntry*) NULL;
       h = Tcl_NextHashEntry (&hs)) {

    pa = (PiTclPackInfo*) Tcl_GetHashValue (h);

    LIST_ADDSTR (list, pa->name);
  }

  Tcl_SetObjResult (interp, list);
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackAdd --
 *
 *	------------------------------------------------*
 *	Add a packer to the packer table.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackAdd (p, interp, pa)
PiTclPackers*         p;
Tcl_Interp*           interp;
PiTclPackInfo* pa;
{
  int            new;
  Tcl_HashEntry* hPtr;

  hPtr = Tcl_CreateHashEntry (&p->cvt, pa->name, &new);

  if (hPtr == (Tcl_HashEntry*) NULL) {
    PiTclPackRelease (pa);
    return TCL_ERROR;
  }

  PiTclPackPreserve (pa);
  Tcl_SetHashValue  (hPtr, pa);
  pa->hPtr = hPtr;
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclPackRemove --
 *
 *	------------------------------------------------*
 *	Removes the specified entry from the registry
 *	of packers.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPackRemove (p, interp, pa)
PiTclPackers*         p;
Tcl_Interp*           interp;
PiTclPackInfo* pa;
{
    Tcl_DeleteHashEntry (pa->hPtr);

    pa->hPtr = (Tcl_HashEntry*) NULL;

    PiTclPackRelease    (pa);

    return TCL_OK;
}

