/*
 * wrap.c --
 *
 *	Implementation of wrappers for the existing C-level packers.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: wrap.c,v 1.1 1999/03/27 13:26:35 aku Exp $
 */


#include "macros.h"
#include "wrap.h"

/* Forward declaration of internal procedures.
 */

static int
HandleStructure _ANSI_ARGS_ ((Tcl_Interp*    interp,
			      PiTcl_CvtData* cvt,
			      int            direction,
			      Tcl_Obj*       in));


/*
 *------------------------------------------------------*
 *
 *	PiTclPilotCvtCmd --
 *
 *	------------------------------------------------*
 *	This procedure realizes the various wrapper
 *	commands, under control of a conversion descriptor.
 *	See the manpages for details on what it does.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *		As of the called conversion functions.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclPilotCvtCmd (cd, interp, objc, objv)
ClientData     cd;       /* <= PiTcl_PilotCvt */
Tcl_Interp*    interp;
int            objc;
Tcl_Obj* CONST objv [];
{
  /* Convert subcommand to easy-to-use index. This converts the
   * given Tcl_Obj too, caching the conversion result. Later
   * accesses are done in constant time :-)
   */

  PiTcl_PilotCvt* cvt = (PiTcl_PilotCvt*) cd;
  int             sub, dir;
  char*           cmd = ARG (0);

  static char *subcmd [] = {
    "appinfo", "record", "sortinfo", NULL
  };
  enum subcmd {
    OP_APPINFO, OP_RECORD, OP_SORTINFO
  };	  

  static char *direction [] = {
    "pack", "unpack", NULL
  };
  enum direction {
    DIR_PACK, DIR_UNPACK
  };	  

  DLP_CHECKARGS (objc < 4, interp, cmd, " <subcommand> <direction> <data>");

  if (Tcl_GetIndexFromObj(interp, objv [1], subcmd, "subcommand", 0,
			  &sub) != TCL_OK) {
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv [2], direction, "direction", 0,
			  &dir) != TCL_OK) {
    return TCL_ERROR;
  }

  /* Now dispatch to the appropriate subcommands the direction to use.
   */

  switch (sub) {
  case OP_APPINFO:
    if (cvt->appinfo == (PiTcl_CvtData*) NULL) {
      Tcl_AppendResult (interp, "appinfo conversion not available", NULL);
      return TCL_ERROR;
    } else {
      return HandleStructure (interp, cvt->appinfo, dir, objv [3]);
    }

  case OP_SORTINFO:
    if (cvt->sortinfo == (PiTcl_CvtData*) NULL) {
      Tcl_AppendResult (interp, "sortinfo conversion not available", NULL);
      return TCL_ERROR;
    } else {
      return HandleStructure (interp, cvt->sortinfo, dir, objv [3]);
    }

  case OP_RECORD:
    return HandleStructure (interp, cvt->record, dir, objv [3]);

  default:
    /* impossible */
    return TCL_ERROR;
  }

  /* We should not come to this place */
  return TCL_ERROR;
}


/*
 *------------------------------------------------------*
 *
 *	HandleStructure --
 *
 *	------------------------------------------------*
 *	Internal helper, executes the conversion between
 *	bytestreams and tcl lists. 
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates a new object.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
HandleStructure (interp, cvt, direction, in)
Tcl_Interp*    interp;
PiTcl_CvtData* cvt;
int            direction;
Tcl_Obj*       in;
{
#define FREE_STRUCT(s) \
  if (cvt->structFree != (PiTcl_FreeStruct*) NULL) { \
    cvt->structFree (s); \
  } \
  Tcl_Free ((char*) s)

  enum direction {
    DIR_PACK, DIR_UNPACK
  };	  

  if (direction == DIR_PACK) {
    /* 'in' contains a tcl list object convertable into the structure
     *
     * i)   allocate memory for the structure.
     * ii)  convert the list object into it,
     * iii) then pack it into the stream (loop until it fits)
     * iv)  and set the buffer as result of the running interpreter.
     */

    Tcl_Obj* out;
    int      len, res, sz;
    char*    buf;
    void*    structure = Tcl_Alloc (cvt->structSize);

    memset (structure, '\0', cvt->structSize);

    res = TclUtil_StructEncode (interp, cvt->structDef, in, structure);

    if (res != TCL_OK) {
      return res;
    }

    sz  = cvt->streamSize;
    len = 0;

    while (1) {
      buf = Tcl_Alloc (sz);
      len = cvt->structPack (structure, buf, sz);

      if (len > 0) {
	break;
      }

      sz *= 2;
      Tcl_Free (buf);
    }

    /*printf ("wrap:packed, length=%d",len);*/

    out = Tcl_NewStringObj (buf, len);
    Tcl_Free (buf);

    if (out == (Tcl_Obj*) NULL) {
      /* failed */

      FREE_STRUCT (structure);
      Tcl_AppendResult (interp, "could not convert structure", NULL);
      return TCL_ERROR;
    }

    Tcl_SetObjResult (interp, out);
    FREE_STRUCT (structure);
    return TCL_OK;

  } else if (direction == DIR_UNPACK) {
    /* 'in' contains the binary data, the byte stream to unpack.
     *
     * i)   allocate memory for the structure.
     * ii)  unpack the stream into it.
     * iii) convert the structure into a tcl list object.
     * iv)  and set it as result of the running interpreter.
     */

    Tcl_Obj* out;
    int      len, res;
    char*    buf       = Tcl_GetStringFromObj (in, &len);
    void*    structure = Tcl_Alloc (cvt->structSize);

    memset (structure, '\0', cvt->structSize);

    res = cvt->structUnpack (structure, buf, len);

    if (res <= 0) {
      /* unpacking failed */

      FREE_STRUCT (structure);
      Tcl_AppendResult (interp, "could not unpack stream", NULL);
      return TCL_ERROR;
    }

    out = TclUtil_StructDecode (interp, cvt->structDef, structure);

    if (out == (Tcl_Obj*) NULL) {
      /* structure conversion failed */

      FREE_STRUCT (structure);
      return TCL_ERROR;
    }

    Tcl_SetObjResult (interp, out);
  
    FREE_STRUCT (structure);
    return TCL_OK;

  } else {
    Tcl_AppendResult (interp,
		      "internal error, illegal direction",
		      NULL);
    return TCL_ERROR;
  }
}
